#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <time.h>
#include <string.h>
#include <getopt.h>
#include <stdint.h>
#include <unistd.h>
#include <limits.h>
#include <signal.h>
#include <errno.h>
#include "EcatDeviceAPI.h"

#define INPUT_ARG_SIZE 1024
char input_arg[INPUT_ARG_SIZE+1];
uint16_t device_no = 0;
uint8_t should_stop = 0;
uint32_t link_up, slaves_responding, al_states, wc;
uint16_t ecat_2024_slave_no = 3;
uint16_t ecat_2011h_slave_no = 2;
uint8_t run_mode = 0;

static int set_slave_aio_property();
/*****************************************************************************/
static void sig_handler(int sig)
{
	should_stop = 1;
}
/*****************************************************************************/
void show_device_state(uint32_t *link_up, uint32_t *slaves_responding, uint32_t *al_states, uint32_t *wc)
{
	char StrAlState[255];
	if(*al_states == 0)
		sprintf(StrAlState,"INIT");
	else if(*al_states == 0x02)
		sprintf(StrAlState,"PREOP");
	else if(*al_states == 0x04)
		sprintf(StrAlState,"SAFEOP");
	else if(*al_states == 0x08)
		sprintf(StrAlState,"OP");
	else
		sprintf(StrAlState,"Invalid");

	printf("Device State\n");
	printf("\t |_  Slave(s): %u\n", *slaves_responding);
	printf("\t |_ AL States: %s\n", StrAlState);
	printf("\t |_      Link: %s\n", *link_up? "up" : "down");
	printf("\t |_        Wc: %-u\n", *wc);
	printf("\n");
}
/*****************************************************************************/
void show_act_list(void)
{
	printf("/**************action**************/\n");
	printf("[1]  - Set Unipolar 10V\n"
				 "[2]  - Set Bipolar 10V\n"
			);
}
/*****************************************************************************/
void handle_act(void)
{
	uint32_t act_no = atoi(input_arg);
	switch(act_no)
	{
	case 1:
		run_mode = 0;
		break;
	case 2:
		run_mode = 1;
		break;
	default:
		printf("Invalid action number\n");
		break;
	}

	if (set_slave_aio_property()) {
		fgets(input_arg, INPUT_ARG_SIZE , stdin);
	}
}
/*****************************************************************************/
int __sleep(void)
{
	struct timeval tv;
	fd_set readfds;
	int syscall_ret;
	ssize_t read_ret;
	tv.tv_sec = 0;
	tv.tv_usec = 1000;
	FD_ZERO(&readfds);
	FD_SET(STDIN_FILENO, &readfds);
	syscall_ret = select(STDIN_FILENO+1, &readfds, NULL, NULL, &tv);
	if(!syscall_ret) {

		return syscall_ret;

	} else {
		if(FD_ISSET(STDIN_FILENO,&readfds)) {
			if (errno != EINTR) {
				read_ret = read(STDIN_FILENO, input_arg, INPUT_ARG_SIZE);
				if(read_ret) {
					show_act_list();
					if (fgets(input_arg, INPUT_ARG_SIZE , stdin) != NULL)
						handle_act();
				}
			}
		}
	}
	return syscall_ret;
}
/*****************************************************************************/
int set_slave_aio_property()
{
	int32_t ret;
	uint16_t i;
	uint8_t ao_range = SLAVE_AO_BI_10V;
	uint8_t ai_range = SLAVE_AO_BI_10V;

	switch (run_mode) {
	case 0:
		ao_range = SLAVE_AO_UNI_10V;
		ai_range = SLAVE_AI_UNI_10V;
		break;
	case 1:
		ao_range = SLAVE_AO_BI_10V;
		ai_range = SLAVE_AI_BI_10V;
		break;
	default:
		ao_range = SLAVE_AO_BI_10V;
		ai_range = SLAVE_AI_BI_10V;
		break;
	}


	for (i = 0; i < 4; i++) {
		ret = ECAT_SetSlaveAoProperty(device_no, ecat_2024_slave_no, i, ao_range);
		if(ret != 0) {
			printf("Failed to set slave[%u] channel%u AO property:%d\n", ecat_2024_slave_no, i, ret);
			return -1;
		}

		ret = ECAT_SetSlaveAiProperty(device_no, ecat_2011h_slave_no, i, ai_range);
		if(ret != 0) {
			printf("Failed to set slave[%u] channel%u AI property:%d\n", ecat_2011h_slave_no, i, ret);
			return -1;
		}
	}

	return 0;
}
/*****************************************************************************/
int main(void) {

	int32_t ret;
	int32_t ao_time_tick = 0;
	int32_t info_time_tick = 0;
	double ao_value[4] = {0, 0.25, 0.5, 0.75};
	double double_value = 0;
	uint16_t i;

	if (signal(SIGINT, sig_handler) == SIG_ERR){
		printf("Failed to signal handler\n");
		return EXIT_FAILURE;
	}

	ret = ECAT_OpenDevice(device_no);
	if(ret != 0) {
		printf("Failed to open device:%d\n", ret);
		return EXIT_FAILURE;
	}



	ret = ECAT_StartDeviceOpTask(device_no, 0, DEV_OP_CYCLE_TIME_1MS, 3);
	if(ret != 0) {
		printf("Failed to start device OP task:%d\n", ret);
		goto out_error;
	}

	set_slave_aio_property();



	while(!should_stop) 	{
		if (__sleep() == 0) {

			ret = ECAT_GetDeviceState(device_no, &link_up, &slaves_responding, &al_states, &wc);

			if (!ret) {
				if (al_states == 0x08) {
					if (ao_time_tick)
						ao_time_tick--;
					else {
						for (i = 0; i < 4; i++) {
							if (ao_value[i] > 10.0) {
								if (run_mode)
									ao_value[i] = -10.0 + (0.25 * i);
								else
									ao_value[i] = (0.25 * i);
							}

							ret = ECAT_SetSlaveAoVoltData(device_no, ecat_2024_slave_no, i, ao_value[i]);
							if (ret) {
								printf("Failed to set slave[%u] channel%u AO vlot:%d\n", ecat_2011h_slave_no, i, ret);
								fgets(input_arg, INPUT_ARG_SIZE , stdin);
							}

							ao_value[i] += 0.25;
						}

						ao_time_tick = 1000;
					}
				}


				if (info_time_tick)
					info_time_tick--;
				else {
					printf("\033[2J"); //clear
					printf("\033[0;0H");
					show_device_state(&link_up, &slaves_responding, &al_states, &wc);


					printf("ECAT-2011H[%u]\n", ecat_2011h_slave_no);
					for (i = 0; i < 4; i++) {
						ret = ECAT_GetSlaveAiVoltData(device_no, ecat_2011h_slave_no, i, &double_value);
						if (!ret)
							printf("\t |_ AI[%u]: %f volt\n", i, double_value);
						else
							printf("\t |_ AI[%u]: ?? volt\n", i);
					}
					printf("\n");

					printf("ECAT-2024[%u]\n", ecat_2024_slave_no);
					for (i = 0; i < 4; i++) {
						ret = ECAT_GetSlaveAoVoltData(device_no, ecat_2024_slave_no, i, &double_value);
						if (!ret)
							printf("\t |_ AO[%u]: %f volt\n", i, double_value);
						else
							printf("\t |_ AO[%u]: ?? volt\n", i);
					}
					printf("\n");



					info_time_tick = 500;
				}
			}
		}
	}

out_error:
	ret = ECAT_CloseDevice(device_no);
	if(ret != 0) {
		printf("Failed to close device:%d\n", ret);
		return EXIT_FAILURE;
	} else
		printf("Close device\n");
	return EXIT_SUCCESS;
}
