#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <time.h>
#include <string.h>
#include <getopt.h>
#include <stdint.h>
#include <unistd.h>
#include <limits.h>
#include <signal.h>
#include <errno.h>
#include "EcatDeviceAPI.h"

#define INPUT_ARG_SIZE 1024
char input_arg[INPUT_ARG_SIZE+1];
uint16_t device_no = 0;
uint8_t should_stop = 0;
uint32_t link_up, slaves_responding, al_states, wc;

/*****************************************************************************/
static void sig_handler(int sig)
{
	should_stop = 1;
}
/*****************************************************************************/
void show_device_state(uint32_t *link_up, uint32_t *slaves_responding, uint32_t *al_states, uint32_t *wc)
{
	char StrAlState[255];
	if(*al_states == 0)
		sprintf(StrAlState,"INIT");
	else if(*al_states == 0x02)
		sprintf(StrAlState,"PREOP");
	else if(*al_states == 0x04)
		sprintf(StrAlState,"SAFEOP");
	else if(*al_states == 0x08)
		sprintf(StrAlState,"OP");
	else
		sprintf(StrAlState,"Invalid");

	printf("Device State\n");
	printf("\t |_  Slave(s): %u\n", *slaves_responding);
	printf("\t |_ AL States: %s\n", StrAlState);
	printf("\t |_      Link: %s\n", *link_up? "up" : "down");
	printf("\t |_        Wc: %-u\n", *wc);
	printf("\n");
}
/*****************************************************************************/
void show_act_list(void)
{
	printf("/**************action**************/\n");

}
/*****************************************************************************/
void handle_act(void)
{
	uint32_t act_no = atoi(input_arg);
	switch(act_no)
	{
	case 1:
		break;
	default:
		printf("Invalid action number\n");
		break;
	}
}
/*****************************************************************************/
int __sleep(void)
{
	struct timeval tv;
	fd_set readfds;
	int syscall_ret;
	ssize_t read_ret;
	tv.tv_sec = 0;
	tv.tv_usec = 1000;
	FD_ZERO(&readfds);
	FD_SET(STDIN_FILENO, &readfds);
	syscall_ret = select(STDIN_FILENO+1, &readfds, NULL, NULL, &tv);
	if(!syscall_ret) {

		return syscall_ret;

	} else {
		if(FD_ISSET(STDIN_FILENO,&readfds)) {
			if (errno != EINTR) {
				read_ret = read(STDIN_FILENO, input_arg, INPUT_ARG_SIZE);
				if(read_ret) {
					show_act_list();
					if (fgets(input_arg, INPUT_ARG_SIZE , stdin) != NULL)
						handle_act();
				}
			}
		}
	}
	return syscall_ret;
}
/*****************************************************************************/
void show_slave_dio_info(char *title, int32_t error, uint32_t *do_value, uint32_t *di_value)
{
	uint16_t i;
	uint32_t u32_value;
	if (title) {
		printf("%s\n", title);
		printf("\t|_  Bit\t 0 1 2 3 4 5 6 7\n");
	}

	if (do_value ) {
		if (!error) {
			u32_value = *do_value;
			printf("\t |_  DO\t ");
			for (i = 0; i < 8; i++) {
				if (u32_value & (0x01 << i))
					printf("1 ");
				else
					printf("0 ");
			}
			printf("\n");
		} else
			printf("\t |_  DO\t ? ? ? ? ? ? ? ?\n");
	}

	if (di_value) {
		if (!error) {
			u32_value = *di_value;
			printf("\t |_  DI\t ");
			for (i = 0; i < 8; i++) {
				if (u32_value & (0x01 << i))
					printf("1 ");
				else
					printf("0 ");
			}
			printf("\n");
		} else
			printf("\t |_  DI\t ? ? ? ? ? ? ? ?\n");
	}
}
/*****************************************************************************/
int main(void) {

	int32_t ret;
	int32_t do_time_tick = 0;
	int32_t info_time_tick = 0;
	uint32_t do_value = 0x55;
	uint32_t u32_value;
	uint16_t ecat_2060_slave_no = 4;
	uint16_t ecat_2055_slave_no = 5;
	char title[512];

	if (signal(SIGINT, sig_handler) == SIG_ERR){
		printf("Failed to signal handler\n");
		return EXIT_FAILURE;
	}

	ret = ECAT_OpenDevice(device_no);
	if(ret != 0) {
		printf("Failed to open device:%d\n", ret);
		return EXIT_FAILURE;
	}

	ret = ECAT_StartDeviceOpTask(device_no, 0, DEV_OP_CYCLE_TIME_1MS, 3);
	if(ret != 0) {
		printf("Failed to start device OP task:%d\n", ret);
		goto out_error;
	}

	while(!should_stop) {
		if (__sleep() == 0) {

			ret = ECAT_GetDeviceState(device_no, &link_up, &slaves_responding, &al_states, &wc);

			if (!ret) {

				if  (al_states == 0x08) {
					if (do_time_tick)
						do_time_tick--;
					else {
						ret = ECAT_SetSlaveDO(device_no, ecat_2060_slave_no, do_value);
						if (ret) {
							printf("Failed to set slave[%u] do:%d\n", ecat_2060_slave_no, ret);
							fgets(input_arg, INPUT_ARG_SIZE , stdin);
						}
						ret = ECAT_SetSlaveDO(device_no, ecat_2055_slave_no, do_value);
						if (ret) {
							printf("Failed to set slave[%u] do:%d\n", ecat_2060_slave_no, ret);
							fgets(input_arg, INPUT_ARG_SIZE , stdin);
						}
						do_value = ~do_value;
						do_time_tick = 3000;
					}
				}


				if (info_time_tick)
					info_time_tick--;
				else {
					printf("\033[2J"); //clear
					printf("\033[0;0H");
					show_device_state(&link_up, &slaves_responding, &al_states, &wc);


					sprintf(title, "ECAT-2060[%u]", ecat_2060_slave_no);
					ret = ECAT_GetSlaveDO(device_no, ecat_2060_slave_no, &u32_value);
					show_slave_dio_info(title, ret, &u32_value, NULL);

					ret = ECAT_GetSlaveDI(device_no, ecat_2060_slave_no, &u32_value);
					show_slave_dio_info(NULL, ret, NULL, &u32_value);
					printf("\n");

					sprintf(title, "ECAT-2055[%u]", ecat_2055_slave_no);
					ret = ECAT_GetSlaveDO(device_no, ecat_2055_slave_no, &u32_value);
					show_slave_dio_info(title, ret, &u32_value, NULL);

					ret = ECAT_GetSlaveDI(device_no, ecat_2055_slave_no, &u32_value);
					show_slave_dio_info(NULL, ret, NULL, &u32_value);

					info_time_tick = 500;
				}
			}
		}
	}

out_error:
	ret = ECAT_CloseDevice(device_no);
	if(ret != 0) {
		printf("Failed to close device:%d\n", ret);
		return EXIT_FAILURE;
	} else
		printf("Close device\n");
	return EXIT_SUCCESS;
}
