#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <time.h>
#include <string.h>
#include <getopt.h>
#include <stdint.h>
#include <unistd.h>
#include <limits.h>
#include <signal.h>
#include <errno.h>
#include "EcatDeviceAPI.h"

#define INPUT_ARG_SIZE 1024
#define AXIS_LOOP_POS1 4
#define AXIS_LOOP_POS2 -5
#define AXIS_LOOP_VEL 2.5

char input_arg[INPUT_ARG_SIZE+1];
uint16_t device_no = 0;
uint32_t link_up, slaves_responding, al_states, wc;
uint16_t servo_motor_slave_no = 1;
uint16_t mc_task_idx = 0;
uint16_t temp_task_idx = 0;
uint16_t mc_task_time_out;
uint8_t mc_task_error;
uint8_t should_stop = 0;

char mc_error_msg[256] = "";
uint8_t run_mode = 0;

/*****************************************************************************/
static void sig_handler(int sig)
{
	should_stop = 1;
}
/*****************************************************************************/
void show_device_state(uint32_t *link_up, uint32_t *slaves_responding, uint32_t *al_states, uint32_t *wc)
{
	char StrAlState[255];

	if(*al_states == 0)
		sprintf(StrAlState,"INIT");
	else if(*al_states == 0x02)
		sprintf(StrAlState,"PREOP");
	else if(*al_states == 0x04)
		sprintf(StrAlState,"SAFEOP");
	else if(*al_states == 0x08)
		sprintf(StrAlState,"OP");
	else
		sprintf(StrAlState,"Invalid");


	printf("Device State\n");
	printf("\t |_  Slave(s): %u\n", *slaves_responding);
	printf("\t |_ AL States: %s\n", StrAlState);
	printf("\t |_      Link: %s\n", *link_up? "up" : "down");
	printf("\t |_        Wc: %-u\n", *wc);
	printf("\n");
}
/*****************************************************************************/
void show_act_list(void)
{
	printf("/**************action**************/\n");
	printf("[1]  - Axis Error Reset\n"
				 "[2]  - Axis Loop Restart\n"
			);
}
/*****************************************************************************/
void handle_act(void)
{
//	int32_t ret;
	uint32_t act_no = atoi(input_arg);
	switch(act_no)
	{
	case 1:
		memset(mc_error_msg, 0, sizeof(mc_error_msg));
		temp_task_idx = mc_task_idx;
		mc_task_idx = 1000;
		mc_task_error = 0;
		break;
	case 2:
		memset(mc_error_msg, 0, sizeof(mc_error_msg));
		mc_task_idx = 0;
		mc_task_error = 0;
		break;
	default:
		printf("Invalid action number\n");
		break;
	}


}
/*****************************************************************************/
int __sleep(void)
{
	struct timeval tv;
	fd_set readfds;
	int syscall_ret;
	ssize_t read_ret;
	tv.tv_sec = 0;
	tv.tv_usec = 1000;
	FD_ZERO(&readfds);
	FD_SET(STDIN_FILENO, &readfds);
	syscall_ret = select(STDIN_FILENO+1, &readfds, NULL, NULL, &tv);
	if(!syscall_ret) {

		return syscall_ret;

	} else {
		if(FD_ISSET(STDIN_FILENO,&readfds)) {
			if (errno != EINTR) {
				read_ret = read(STDIN_FILENO, input_arg, INPUT_ARG_SIZE);
				if(read_ret) {
					show_act_list();
					if (fgets(input_arg, INPUT_ARG_SIZE , stdin) != NULL)
						handle_act();
				}
			}
		}
	}
	return syscall_ret;
}
/*****************************************************************************/
int32_t check_axis_standstill()
{
	int32_t ret;
	uint32_t u32_value;
	ret = ECAT_McGetAxisState(device_no, 0, &u32_value);
	if (!ret) {
		if (u32_value == MC_AS_STANDSTILL)
			return 0;
		else if (u32_value == MC_AS_ERRORSTOP)
			return -1;
		else
			return 1;
	} else {
		sprintf(mc_error_msg, "Failed to get axis0 state:%d", ret);
		mc_task_error = 1;
		return -1;
	}
}
/*****************************************************************************/
void mc_task(void)
{
	int32_t ret;
	uint32_t u32_value;

	switch(mc_task_idx) {
		case 0:
			ret = ECAT_McSetAxisServoOn(device_no, 0, 1);
			if (ret) {
				sprintf(mc_error_msg, "Failed to set axis0 Servo ON:%d", ret);
				mc_task_error = 1;
			} else {
				mc_task_idx++;
				mc_task_time_out = 500;
			}
			break;
		case 1:
			ret = check_axis_standstill();
			if (ret == 0)
				mc_task_idx++;
			else if (ret == -1) {
				if (mc_task_error)
					return;

				if (mc_task_time_out)
					mc_task_time_out--;
				else {
					sprintf(mc_error_msg, "Invalid axis state");
					mc_task_error = 1;
				}
			}
			break;
		case 2:
			ret = ECAT_McGetAxisDI(device_no, 0, &u32_value);
			if (!ret) {
				if (u32_value & 0x20)
					mc_task_idx++;
			}
			break;
		case 3:
			ret = ECAT_McAxisMoveRel(device_no, 0, AXIS_LOOP_POS1, AXIS_LOOP_VEL);
			if (ret) {
				sprintf(mc_error_msg, "Axis0 move relative failed:%d", ret);
				mc_task_error = 1;
			} else {
				mc_task_idx++;
			}
			break;
		case 4:
			ret = check_axis_standstill();
			if (ret == 0)
				mc_task_idx++;
			else if (ret == -1) {
				if (mc_task_error)
					return;

				sprintf(mc_error_msg, "Invalid axis state");
				mc_task_error = 1;
			}
			break;
		case 5:
			ret = ECAT_McAxisMoveRel(device_no, 0, AXIS_LOOP_POS2, AXIS_LOOP_VEL);
			if (ret) {
				sprintf(mc_error_msg, "Axis0 move relative failed:%d", ret);
				mc_task_error = 1;
			} else {
				mc_task_idx++;
			}
			break;
		case 6:
			ret = check_axis_standstill();
			if (ret == 0)
				mc_task_idx = 1;
			else if (ret == -1) {
				if (mc_task_error)
					return;

				sprintf(mc_error_msg, "Invalid axis state");
				mc_task_error = 1;
			}
			break;
		case 1000:
			ret = ECAT_McAxisErrorReset(device_no, 0);
			if (ret) {
				sprintf(mc_error_msg, "Axis0 error reset failed:%d", ret);
				mc_task_error = 1;
			} else {
				mc_task_idx++;
				mc_task_time_out = 100;
			}
			break;
		case 1001:
			ret = check_axis_standstill();
			if (ret == 0)
				mc_task_idx = temp_task_idx;
			else if (ret == -1) {
				if (mc_task_error)
					return;
				if (mc_task_time_out)
					mc_task_time_out--;
				else {
					sprintf(mc_error_msg, "Axis error reset timeout");
					mc_task_error = 1;
				}
			}
			break;

		default:
			break;
	}
}
/*****************************************************************************/
int main(void) {

	int32_t ret;
	int32_t info_time_tick = 0;
	double double_value = 0;
	uint32_t u32_value;
	int32_t i32_value;
	uint16_t u16_value;
	char axis_state[128];
	uint16_t slave_no[3];
	uint16_t sub_axis_no[3];
	uint16_t axis_cnt;
	uint16_t i;

	if (signal(SIGINT, sig_handler) == SIG_ERR){
		printf("Failed to signal handler\n");
		return EXIT_FAILURE;
	}


	ret = ECAT_OpenDevice(device_no);
	if(ret != 0) {
		printf("Failed to open device:%d\n", ret);
		return EXIT_FAILURE;
	}


	ret = ECAT_StartDeviceOpTask(device_no, 0, DEV_OP_CYCLE_TIME_1MS, 3);
	if(ret != 0) {
		printf("Failed to start device OP task:%d\n", ret);
		goto out_error;
	}

	slave_no[0] = servo_motor_slave_no;
	sub_axis_no[0] = 0;
	axis_cnt = 1;

	ret = ECAT_McInit(device_no, slave_no, sub_axis_no, axis_cnt);
	if(ret != 0) {
		printf("Failed to init motion:%d\n", ret);
		goto out_error;
	}

	ret = ECAT_McSetAxisPPU(device_no, 0, 100000);
	if(ret != 0) {
		printf("Failed to set axis0 PPU:%d\n", ret);
		fgets(input_arg, INPUT_ARG_SIZE , stdin);
	}


	while(!should_stop) {
		if (__sleep() == 0) {

			ret = ECAT_GetDeviceState(device_no, &link_up, &slaves_responding, &al_states, &wc);

			if ((!ret)) {

				if (al_states == 0x08) {
					if (!mc_task_error)
						mc_task();
				}

				if (info_time_tick)
					info_time_tick--;
				else {
					printf("\033[2J"); //clear
					printf("\033[0;0H");

					show_device_state(&link_up, &slaves_responding, &al_states, &wc);


					printf("Axis0[%u]-Servo\n", servo_motor_slave_no);

					ret = ECAT_McGetAxisActualPos(device_no, 0, &double_value);
					if (!ret)
						printf("\t |_ Position: %f\n", double_value);
					else
						printf("\t |_ Position: ??\n");

					ret = ECAT_McGetAxisActualVel(device_no, 0, &double_value);
					if (!ret)
						printf("\t |_ Velocity: %f\n", double_value);
					else
						printf("\t |_ Velocity: ??\n");

					ret = ECAT_McGetAxisState(device_no, 0, &u32_value);
					if (!ret) {
						switch(u32_value) {
								case MC_AS_DISABLED:
									strcpy(axis_state, "Disabled");
									break;
								case MC_AS_STANDSTILL:
									strcpy(axis_state, "StandStill");
									break;
								case MC_AS_ERRORSTOP:
									strcpy(axis_state, "ErrorStop");
									break;
								case MC_AS_STOPPING:
									strcpy(axis_state, "Stopping");
									break;
								case MC_AS_HOMING:
									strcpy(axis_state, "Homing");
									break;
								case MC_AS_DISCRETEMOTION:
									strcpy(axis_state, "DiscMotion");
									break;
								case MC_AS_CONTINUOUSMOTION:
									strcpy(axis_state, "ContMotion");
									break;
								case MC_AS_SYNCHRONIZEDMOTION:
									strcpy(axis_state, "SyncMotion");
									break;
								default:
									strcpy(axis_state, "Invalid");
									break;
							}
						printf("\t |_    State: %s\n", axis_state);
					} else
						printf("\t |_    State: ??\n");

					ret =  ECAT_McGetAxisLastError(device_no, 0, &i32_value);
					if (!ret)
						printf("\t |_    Error: %d,", i32_value);
					else
						printf("\t |_    Error: ??,");

					ret = ECAT_McGetAxisDriveError(device_no, 0, &u16_value);
					if (!ret)
						printf("%u\n", u16_value);
					else
						printf("??\n");

					ret = ECAT_McGetAxisDI(device_no, 0, &u32_value);
					if (!ret) {
						printf("\t |_       DI: ");
						for (i = 0; i < 7; i++) {
							if (u32_value & (0x01 << i))
								printf("1 ");
							else
								printf("0 ");
						}
						printf("\n");
					}
					else
						printf("\t |_       DI: ??\n");
					printf("\n");


					printf("Motion Control Task\n");
					printf("\t |_ Index: %u\n", mc_task_idx);
					printf("\t |_ Error: %s\n", mc_error_msg);

					info_time_tick = 500;
				}
			}
		}
	}

	ECAT_McAxisQuickStop(device_no, 0);
	usleep(100000);

out_error:
	ret = ECAT_CloseDevice(device_no);
	if(ret != 0) {
		printf("Failed to close device:%d\n", ret);
		return EXIT_FAILURE;
	} else
		printf("Close device\n");
	return EXIT_SUCCESS;
}
